@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 2000-2013  Ulric Eriksson <ulric@siag.nu>
+   Copyright (C) 2000-2014  Ulric Eriksson <ulric@siag.nu>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -38,6 +38,7 @@
 #include <sys/wait.h>
 #include <sys/resource.h>
 #include <sys/socket.h>
+#include <sys/un.h>
 #include <sys/stat.h>
 #ifdef HAVE_POLL
 #include <sys/poll.h>
@@ -70,6 +71,15 @@ static char *cacert_file;
 static SSL_CTX *ssl_context = NULL;
 #endif  /* HAVE_SSL */
 
+#ifdef HAVE_GEOIP
+#include <GeoIP.h>
+GeoIP *geoip;
+#endif
+
+#define ACE_IPV4 (1)
+#define ACE_IPV6 (2)
+#define ACE_GEO (3)
+
 #define BUFFER_MAX 	(32*1024)
 
 #define CLIENTS_MAX	2048	/* max clients */
@@ -120,7 +130,29 @@ typedef struct {
 
 typedef struct {
 	unsigned int ip, mask;
+} ace_ipv4;
+
+typedef struct {
+	char ip[INET6_ADDRSTRLEN];
+	unsigned char len;
+} ace_ipv6;
+
+typedef struct {
+	char country[2];
+} ace_geo;
+
+typedef struct {
+	unsigned char class;
+	unsigned char permit;
+	union {
+		ace_ipv4 ipv4;
+		ace_ipv6 ipv6;
+		ace_geo geo;
+	} ace;
+#if 0
+	unsigned int ip, mask;
 	int permit;
+#endif
 } acl;
 
 static int nservers;		/* number of servers */
@@ -179,7 +211,7 @@ static FILE *pidfp = NULL;
 static char *webfile = NULL;
 static char *listenport = NULL;
 static char *ctrlport = NULL;
-static int listenfd, ctrlfd;
+static int listenfd, ctrlfd = -1;
 static char *e_server = NULL;
 static char *a_server = NULL;
 static char *jail = NULL;
@@ -430,22 +462,59 @@ static int ssl_init(void)
 
 #endif  /* HAVE_SSL */
 
-static void add_acl(int a, unsigned int ip, unsigned int mask, int permit)
+/* allocate ace and fill in the generics */
+static int add_acl(int a, unsigned char permit)
 {
 	int i;
-
-	if (debuglevel) {
-		debug("add_acl(%d, %x, %x, %d)", a, ip, mask, permit);
-	}
 	if (a < 0 || a >= ACLS_MAX) {
 		debug("add_acl: %d outside (0,%d)", a, ACLS_MAX);
-		return;
+		return -1;
 	}
 	i = nacls[a]++;
 	acls[a] = pen_realloc(acls[a], nacls[a]*sizeof(acl));
-	acls[a][i].ip = ip;
-	acls[a][i].mask = mask;
 	acls[a][i].permit = permit;
+	return i;
+}
+
+static void add_acl_ipv4(int a, unsigned int ip, unsigned int mask, unsigned char permit)
+{
+	int i = add_acl(a, permit);
+
+	if (i == -1) return;
+
+	if (debuglevel) {
+		debug("add_acl_ipv4(%d, %x, %x, %d)", a, ip, mask, permit);
+	}
+	acls[a][i].class = ACE_IPV4;
+	acls[a][i].ace.ipv4.ip = ip;
+	acls[a][i].ace.ipv4.mask = mask;
+}
+
+static void add_acl_ipv6(int a, char *ipaddr, unsigned char len, unsigned char permit)
+{
+	int i = add_acl(a, permit);
+
+	if (i == -1) return;
+
+	if (debuglevel) {
+		debug("add_acl_ipv6(%d, %x, %d, %d)", a, ipaddr, len, permit);
+	}
+	acls[a][i].class = ACE_IPV6;
+	memcpy(acls[a][i].ace.ipv6.ip, ipaddr, INET6_ADDRSTRLEN);
+	acls[a][i].ace.ipv6.len = len;
+}
+
+static void add_acl_geo(int a, char *country, unsigned char permit)
+{
+	int i = add_acl(a, permit);
+
+	if (i == -1) return;
+
+	if (debuglevel) {
+		debug("add_acl_geo(%d, %s, %d", a, country, permit);
+	}
+	acls[a][i].class = ACE_GEO;
+	strncpy(acls[a][i].ace.geo.country, country, 2);
 }
 
 static void del_acl(int a)
@@ -462,25 +531,53 @@ static void del_acl(int a)
 	nacls[a] = 0;
 }
 
-static int match_acl(int a, unsigned int client)
+/* client is an ipv4 address; to be continued... */
+static int match_acl_ipv4(int a, unsigned int client)
 {
 	int i;
 	int permit = 0;
 	acl *ap = acls[a];
+	const char *country = NULL;
+	struct in_addr *in = (struct in_addr *)&client;
 	if (debuglevel) {
-		debug("match_acl(%d, %u)", a, client);
+		debug("match_acl_ipv4(%d, %u)", a, client);
 	}
 	for (i = 0; i < nacls[a]; i++) {
-		if ((client & ap[i].mask) == ap[i].ip) {
-			return ap[i].permit;
-		} else {
-			permit = ap[i].permit;
+		permit = ap[i].permit;
+		switch (ap[i].class) {
+		case ACE_IPV4:
+			if ((client & ap[i].ace.ipv4.mask) == ap[i].ace.ipv4.ip) {
+				return permit;
+			}
+			break;
+		case ACE_IPV6:
+			debug("ACE_IPV6: Not implemented");
+			break;
+		case ACE_GEO:
+#ifdef HAVE_GEOIP
+			if (geoip == NULL) break;
+			if (country == NULL) {
+				country = GeoIP_country_code_by_addr(geoip,
+						inet_ntoa(*in));
+			}
+			if (country && !strncmp(country,
+						ap[i].ace.geo.country, 2)) {
+				return permit;
+			}
+#else
+			debug("ACE_GEO: Not implemented");
+#endif
+			break;
+		default:
+			debug("Unknown ACE class %d (this is probably a bug)",
+				ap[i].class);
+			break;
 		}
 	}
 	return !permit;
 }
 
-static void webstats(void)
+static int webstats(void)
 {
 	FILE *fp;
 	int i;
@@ -488,8 +585,15 @@ static void webstats(void)
 	struct tm *nowtm;
 	char nowstr[80];
 
+	if (webfile == NULL) {
+		debug("Don't know where to write web stats; see -w option");
+		return 0;
+	}
 	fp = fopen(webfile, "w");
-	if (fp == NULL) return;
+	if (fp == NULL) {
+		debug("Can't write to %s", webfile);
+		return 0;
+	}
 	now=time(NULL);
 	nowtm = localtime(&now);
 	strftime(nowstr, sizeof(nowstr), "%Y-%m-%d %H:%M:%S", nowtm);
@@ -607,6 +711,7 @@ static void webstats(void)
 		"</body>\n"
 		"</html>\n");
 	fclose(fp);
+	return 1;
 }
 
 static void textstats(void)
@@ -1006,7 +1111,7 @@ static int copy_down(int i)
 
 	unsigned char b[BUFFER_MAX];
 
-	size_t size = sizeof(struct sockaddr);
+	socklen_t size = sizeof(struct sockaddr);
 
 	if (!udp)
 		rc = read(from, b, BUFFER_MAX);	/* no ssl here */
@@ -1397,7 +1502,7 @@ static int try_server(int index, int sticky, unsigned int cli_addr,
 				servers[index].maxc, servers[index].hard);
 		return -1;
 	}
-	if (!match_acl(servers[index].acl, cli_addr)) {
+	if (!match_acl_ipv4(servers[index].acl, cli_addr)) {
 		if (debuglevel) debug("try_server: denied by acl");
 		return -1;
 	}
@@ -1495,12 +1600,25 @@ static void write_cfg(char *p)
 	for (i = 0; i < ACLS_MAX; i++) {
 		fprintf(fp, "no acl %d\n", i);
 		for (j = 0; j < nacls[i]; j++) {
-			memcpy(&ip, &acls[i][j].ip, 4);
-			fprintf(fp, "acl %d %s %s ", i,
-				acls[i][j].permit?"permit":"deny",
-				inet_ntoa(ip));
-			memcpy(&ip, &acls[i][j].mask, 4);
-			fprintf(fp, "%s\n", inet_ntoa(ip));
+			switch (acls[i][j].class) {
+			case ACE_IPV4:
+				memcpy(&ip, &acls[i][j].ace.ipv4.ip, 4);
+				fprintf(fp, "acl %d %s %s ", i,
+					acls[i][j].permit?"permit":"deny",
+					inet_ntoa(ip));
+				memcpy(&ip, &acls[i][j].ace.ipv4.mask, 4);
+				fprintf(fp, "%s\n", inet_ntoa(ip));
+				break;
+			case ACE_IPV6:
+				debug("ACE_IPV6: Not implemented");
+				break;
+			case ACE_GEO:
+				debug("ACE_GEO: Not implemented");
+				break;
+			default:
+				debug("Unknown ACE class %d (this is probably a bug)",
+					acls[i][j].class);
+			}
 		}
 	}
 	if (asciidump) fprintf(fp, "ascii\n");
@@ -1567,24 +1685,46 @@ static void do_cmd(char *b, void (*output)(char *, void *), void *op)
 		    (ip = strtok(NULL, " "))) {
 			int a = atoi(no);
 			int permit;
-			struct in_addr ipaddr, mask;
-			ma = strtok(NULL, " ");
-			if (!ma) ma = "255.255.255.255";
 			if (!strcmp(pd, "permit")) permit = 1;
 			else if (!strcmp(pd, "deny")) permit = 0;
 			else {
 				debug("acl: expected permit|deny");
 				return;
 			}
-			if (!inet_aton(ip, &ipaddr)) {
-				debug("acl: bogus address '%s'\n", ip);
-				return;
+			if (!strcmp(ip, "country")) {
+				char *country = strtok(NULL, " ");
+				if (!country) {
+					debug("acl: expected country");
+					return;
+				}
+				add_acl_geo(a, country, permit);
+			} else if (strchr(ip, ':')) {
+				char ipaddr[INET6_ADDRSTRLEN];
+				ma = strchr(ip, '/');
+				if (ma) {
+					*ma++ = '\0';
+				} else {
+					ma = "128";
+				}
+				if (inet_pton(AF_INET6, ip, ipaddr) != 1) {
+					debug("acl: can't convert address %s", ip);
+					return;
+				}
+				add_acl_ipv6(a, ipaddr, atoi(ma), permit);
+			} else {
+				struct in_addr ipaddr, mask;
+				ma = strtok(NULL, " ");
+				if (!ma) ma = "255.255.255.255";
+				if (!inet_aton(ip, &ipaddr)) {
+					debug("acl: bogus address '%s'\n", ip);
+					return;
+				}
+				if (!inet_aton(ma, &mask)) {
+					debug("acl: bogus mask '%s'\n", ma);
+					return;
+				}
+				add_acl_ipv4(a, ipaddr.s_addr, mask.s_addr, permit);
 			}
-			if (!inet_aton(ma, &mask)) {
-				debug("acl: bogus mask '%s'\n", ma);
-				return;
-			}
-			add_acl(a, ipaddr.s_addr, mask.s_addr, permit);
 		}
 	} else if (!strcmp(p, "ascii")) {
 		asciidump = 1;
@@ -1763,16 +1903,19 @@ static void do_cmd(char *b, void (*output)(char *, void *), void *op)
 			output(b, op);
 		}
 	} else if (!strcmp(p, "status")) {
-		p = webfile;
-		webfile = "/tmp/webfile.html";
-		webstats();
-		fp = fopen(webfile, "r");
-		webfile = p;
-		if (fp == NULL) return;
-		while (fgets(b, sizeof b, fp)) {
-			output(b, op);
+		if (webstats()) {
+			fp = fopen(webfile, "r");
+			if (fp == NULL) {
+				debug("Can't read webstats");
+				return;
+			}
+			while (fgets(b, sizeof b, fp)) {
+				output(b, op);
+			}
+			fclose(fp);
+		} else {
+			debug("Unable to create webstats");
 		}
-		fclose(fp);
 	} else if (!strcmp(p, "stubborn")) {
 		stubborn = 1;
 	} else if (!strcmp(p, "timeout")) {
@@ -1827,7 +1970,7 @@ static void do_ctrl(int downfd, struct sockaddr_in *cli_addr)
 	char b[4096];
 	int n, max_b = sizeof b;
 
-	if (!match_acl(control_acl, cli_addr->sin_addr.s_addr)) {
+	if (!match_acl_ipv4(control_acl, cli_addr->sin_addr.s_addr)) {
 		debug("do_ctrl: not from there");
 	} else {
 		n = read(downfd, b, max_b-1);
@@ -1928,7 +2071,7 @@ static void add_client(int downfd, struct sockaddr_in *cli_addr)
 	}
 #endif
 
-	pd = match_acl(client_acl, cli);
+	pd = match_acl_ipv4(client_acl, cli);
 	if (!pd) {
 		if (debuglevel) debug("add_client: denied by acl");
 		if (abuse_server != -1) {
@@ -2004,6 +2147,25 @@ Success2:
 	return;
 }
 
+static int open_unix_listener(char *a)
+{
+	int listenfd;
+	struct sockaddr_un serv_addr;
+
+	remove(a);
+	memset(&serv_addr, 0, sizeof serv_addr);
+	serv_addr.sun_family = AF_UNIX;
+	snprintf(serv_addr.sun_path, sizeof serv_addr.sun_path, "%s", a);
+	if ((listenfd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
+		error("can't open stream socket");
+	}
+	if (bind(listenfd, (struct sockaddr *)&serv_addr, sizeof serv_addr) != 0) {
+		error("can't bind local address");
+	}
+	listen(listenfd, 50);
+	return listenfd;
+}
+
 static int open_listener(char *a)
 {
 	int listenfd;
@@ -2013,6 +2175,9 @@ static int open_listener(char *a)
 	int fl;
 	int optval = 1;
 
+	/* Handle Unix domain sockets separately */
+	if (strchr(a, '/')) return open_unix_listener(a);
+
 	memset(&serv_addr, 0, sizeof serv_addr);
 	serv_addr.sin_family = AF_INET;
 	p = strchr(a, ':');
@@ -2895,8 +3060,13 @@ int main(int argc, char **argv)
 
 	/* we must open listeners before dropping privileges */
 	/* Control port */
-	if (ctrlport) ctrlfd = open_listener(ctrlport);
-	else ctrlfd = -1;
+	if (ctrlport) {
+		if (getuid() == 0 && user == NULL) {
+			debug("Won't open control port running as root; use -u to run as different user");
+		} else {
+			ctrlfd = open_listener(ctrlport);
+		}
+	}
 
 	/* Balancing port */
 	if (udp) {
@@ -2941,6 +3111,11 @@ int main(int argc, char **argv)
 		fclose(pidfp);
 	}
 
+#ifdef HAVE_GEOIP
+	geoip = GeoIP_new(GEOIP_MEMORY_CACHE);
+	if (geoip == NULL) debug("Could not initialize GeoIP");
+#endif
+
 #ifdef HAVE_KQUEUE
 	if (use_kqueue) mainloop_kqueue();
 #else
